@@ -1,20 +1,24 @@
-// {{{ GPL License
+// {{{ MIT License
 
-// This file is part of gringo - a grounder for logic programs.
-// Copyright Roland Kaminski
+// Copyright 2017 Roland Kaminski
 
-// This program is free software: you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to
+// deal in the Software without restriction, including without limitation the
+// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+// sell copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 
-// You should have received a copy of the GNU General Public License
-// along with this program.  If not, see <http://www.gnu.org/licenses/>.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+// IN THE SOFTWARE.
 
 // }}}
 
@@ -139,7 +143,7 @@ private:
     }
 
     int map_item(std::string &&item) {
-        auto &ret = *item_map_.emplace(std::move(item), item_map_.size()).first;
+        auto &ret = *item_map_.emplace(std::move(item), static_cast<int>(item_map_.size())).first;
         return ret.second;
     }
 
@@ -187,7 +191,7 @@ private:
     }
 
     void initialize_states(PropagateInit &init) {
-        for (int i = 0; i < init.number_of_threads(); ++i) { states_.emplace_back(pattern_length_, sequence_atoms_.size()); }
+        for (int i = 0; i < init.number_of_threads(); ++i) { states_.emplace_back(pattern_length_, static_cast<int>(sequence_atoms_.size())); }
     }
 
     // {{{2 propagation
@@ -241,7 +245,7 @@ public:
         auto &state = states_[ctl.thread_id()];
         uint32_t dl = ctl.assignment().decision_level();
         if (state.trail.size() == 0 || state.trail.back().decision_level < dl) {
-            state.trail.emplace_back(dl, state.stack.size());
+            state.trail.emplace_back(dl, static_cast<int>(state.stack.size()));
         }
         for (auto &lit : changes) {
             for (auto &pat_atom : pattern_atoms_[lit]) {
@@ -424,7 +428,7 @@ private:
     literal_t a_;
     literal_t b_;
     literal_t c_;
-    int count_ = 0;
+    size_t count_ = 0;
 };
 
 class TestAddClause : public Propagator {
@@ -448,9 +452,10 @@ public:
 private:
     literal_t a_;
     literal_t b_;
-    int count_ = 0;
+    size_t count_ = 0;
 };
 
+
 class TestException : public Propagator {
 public:
     void check(PropagateControl &) override {
@@ -458,6 +463,29 @@ public:
     }
 };
 
+class TestMode : public Propagator {
+public:
+    void init(PropagateInit &init) override {
+        auto atoms = init.symbolic_atoms();
+        auto sig = Clingo::Signature("p", 1);
+        for (auto atom : Clingo::make_range(atoms.begin(sig), atoms.end())) {
+            lits_.insert(init.solver_literal(atom.literal()));
+        }
+        init.set_check_mode(Clingo::PropagatorCheckMode::Partial);
+    }
+    void check(PropagateControl &ctl) override {
+        //if (!ctl.assignment().is_total()) { throw std::logic_error("unexpected total check"); }
+        for (auto &lit : lits_) {
+            if (ctl.assignment().truth_value(lit) == Clingo::TruthValue::Free) {
+                ctl.add_clause({lit});
+                break;
+            }
+        }
+    }
+private:
+    std::set<Clingo::literal_t> lits_;
+};
+
 TEST_CASE("propagator", "[clingo][propagator]") {
     MessageVec messages;
     ModelVec models;
@@ -470,12 +498,12 @@ TEST_CASE("propagator", "[clingo][propagator]") {
         auto place = [](int p, int h) { return Function("place", {Number(p), Number(h)}); };
         SECTION("unsat") {
             ctl.ground({{"pigeon", {Number(5), Number(6)}}}, nullptr);
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.empty());
         }
         SECTION("sat") {
             ctl.ground({{"pigeon", {Number(2), Number(2)}}}, nullptr);
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models == ModelVec({{place(1,1), place(2,2)}, {place(1,2), place(2,1)}}));
         }
     }
@@ -484,37 +512,50 @@ TEST_CASE("propagator", "[clingo][propagator]") {
         ctl.register_propagator(p, false);
         ctl.add("base", {}, "{a; b}. c.");
         ctl.ground({{"base", {}}}, nullptr);
-        ctl.solve(MCB(models));
+        test_solve(ctl.solve(), models);
         REQUIRE(models.size() == 4);
     }
+    SECTION("mode") {
+        TestMode prop;
+        ctl.register_propagator(prop, false);
+        ctl.add("base", {}, "{p(1..9)}.");
+        ctl.ground({{"base", {}}}, nullptr);
+        test_solve(ctl.solve(), models);
+        auto p = [](int n) { return Function("p", {Number(n)}); };
+        REQUIRE(models == ModelVec({{ p(1), p(2), p(3), p(4), p(5), p(6), p(7), p(8), p(9) }}));
+    }
     SECTION("exception") {
         TestException p;
         ctl.register_propagator(p, false);
         ctl.add("base", {}, "{a}.");
         ctl.ground({{"base", {}}}, nullptr);
         try {
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             FAIL("solve must throw");
         }
         catch (std::runtime_error const &e) { REQUIRE(e.what() == S("the answer is 42")); }
     }
-#if defined(WITH_THREADS) && WITH_THREADS == 1
     SECTION("exception-t2") {
-        ctl.configuration()["solve.parallel_mode"] = "2";
-        TestException p;
-        ctl.register_propagator(p, false);
-        ctl.add("base", {}, "{a}.");
-        ctl.ground({{"base", {}}}, nullptr);
-        try {
-            ctl.solve(MCB(models));
-            FAIL("solve must throw");
+        bool skip = false;
+        try { ctl.configuration()["solve.parallel_mode"] = "2"; }
+        catch (std::exception const &e) {
+            if (std::strcmp(e.what(), "invalid key") == 0) { skip = true; }
+            else { throw; }
         }
-        catch (std::runtime_error const &e) {
-            // NOTE: in the multithreaded case a runtime error is thrown after the last thread died
-            REQUIRE(e.what() == S("RUNTIME ERROR!"));
+        if (!skip) {
+            TestException p;
+            ctl.register_propagator(p, false);
+            ctl.add("base", {}, "{a}.");
+            ctl.ground({{"base", {}}}, nullptr);
+            try {
+                test_solve(ctl.solve(), models);
+                FAIL("solve must throw");
+            }
+            catch (std::runtime_error const &e) {
+                REQUIRE(e.what() == S("the answer is 42"));
+            }
         }
     }
-#endif
     SECTION("add_clause") {
         TestAddClause p;
         ctl.register_propagator(p, false);
@@ -522,40 +563,40 @@ TEST_CASE("propagator", "[clingo][propagator]") {
             p.type = ClauseType::Learnt;
             ctl.add("base", {}, "{a; b}.");
             ctl.ground({{"base", {}}}, nullptr);
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 3);
             p.enable = false;
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() >= 3);
         }
         SECTION("static") {
             p.type = ClauseType::Static;
             ctl.add("base", {}, "{a; b}.");
             ctl.ground({{"base", {}}}, nullptr);
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 3);
             p.enable = false;
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 3);
         }
         SECTION("volatile") {
             p.type = ClauseType::Volatile;
             ctl.add("base", {}, "{a; b}.");
             ctl.ground({{"base", {}}}, nullptr);
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 3);
             p.enable = false;
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 4);
         }
         SECTION("volatile static") {
             p.type = ClauseType::VolatileStatic;
             ctl.add("base", {}, "{a; b}.");
             ctl.ground({{"base", {}}}, nullptr);
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 3);
             p.enable = false;
-            ctl.solve(MCB(models));
+            test_solve(ctl.solve(), models);
             REQUIRE(models.size() == 4);
         }
     }
@@ -565,7 +606,7 @@ TEST_CASE("propgator-sequence-mining", "[clingo][propagator]") {
     MessageVec messages;
     ModelVec models;
     Logger logger = [&messages](WarningCode code, char const *msg) { messages.emplace_back(code, msg); };
-#if defined(WITH_THREADS) && WITH_THREADS == 1
+#if defined(CLASP_HAS_THREADS) && CLASP_HAS_THREADS == 1
     Control ctl{{"-t14"}, logger, 20};
 #else
     Control ctl{{}, logger, 20};
@@ -618,7 +659,7 @@ TEST_CASE("propgator-sequence-mining", "[clingo][propagator]") {
             }
         }
         int64_t optimum = std::numeric_limits<int64_t>::max();
-        ctl.solve([&models, &optimum](Model m) {
+        for (auto m : ctl.solve()) {
             int64_t opt = m.cost()[0];
             if (opt == optimum) {
                 models.emplace_back();
@@ -628,8 +669,7 @@ TEST_CASE("propgator-sequence-mining", "[clingo][propagator]") {
                 std::sort(models.back().begin(), models.back().end());
             }
             else { optimum = opt; }
-            return true;
-        });
+        }
         std::sort(models.begin(), models.end());
         auto pat = [](int num, char const *item) { return Function("pat", {Number(num), Id(item)}); };
         ModelVec solution = {
