@@ -1,20 +1,24 @@
-// {{{ GPL License
+// {{{ MIT License
 
-// This file is part of gringo - a grounder for logic programs.
-// Copyright (C) 2013  Roland Kaminski
+// Copyright 2017 Roland Kaminski
 
-// This program is free software: you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to
+// deal in the Software without restriction, including without limitation the
+// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+// sell copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 
-// You should have received a copy of the GNU General Public License
-// along with this program.  If not, see <http://www.gnu.org/licenses/>.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+// IN THE SOFTWARE.
 
 // }}}
 
@@ -24,7 +28,9 @@
 #include "gringo/ground/statements.hh"
 #include "gringo/ground/literals.hh"
 #include "gringo/logger.hh"
-
+#ifdef _MSC_VER
+#pragma warning (disable : 4503) // decorated name length exceeded
+#endif
 namespace Gringo { namespace Input {
 
 namespace {
@@ -114,7 +120,7 @@ void warnGlobal(VarTermBoundVec &vars, bool warn, Logger &log) {
         std::sort(ib, ie, [](VarTermBoundVec::value_type const &a, VarTermBoundVec::value_type const &b) { return a.first->name < b.first->name; });
         ie = std::unique(ib, ie, [](VarTermBoundVec::value_type const &a, VarTermBoundVec::value_type const &b) { return a.first->name == b.first->name; });
         for (auto it = ib; it != ie; ++it) {
-            GRINGO_REPORT(log, clingo_warning_global_variable)
+            GRINGO_REPORT(log, Warnings::GlobalVariable)
                 << it->first->loc() << ": info: global variable in tuple of aggregate element:\n"
                 << "  " << it->first->name << "\n"
                 ;
@@ -324,8 +330,8 @@ CreateBody TupleBodyAggregate::toGround(ToGroundArg &x, Ground::UStmVec &stms) c
                 return std::move(ret);
             });
         }
-        return CreateBody([&completeRef, this](Ground::ULitVec &lits, bool primary, bool) {
-            if (primary) { lits.emplace_back(gringo_make_unique<Ground::BodyAggregateLiteral>(completeRef, naf)); }
+        return CreateBody([&completeRef, this](Ground::ULitVec &lits, bool primary, bool auxiliary) {
+            if (primary) { lits.emplace_back(gringo_make_unique<Ground::BodyAggregateLiteral>(completeRef, naf, auxiliary)); }
         }, std::move(split));
     }
     else {
@@ -362,8 +368,8 @@ CreateBody TupleBodyAggregate::toGround(ToGroundArg &x, Ground::UStmVec &stms) c
                 return std::move(ret);
             });
         }
-        return CreateBody([&completeRef, this](Ground::ULitVec &lits, bool primary, bool) {
-            if (primary) { lits.emplace_back(gringo_make_unique<Ground::AssignmentAggregateLiteral>(completeRef)); }
+        return CreateBody([&completeRef, this](Ground::ULitVec &lits, bool primary, bool auxiliary) {
+            if (primary) { lits.emplace_back(gringo_make_unique<Ground::AssignmentAggregateLiteral>(completeRef, auxiliary)); }
         }, std::move(split));
     }
 }
@@ -752,8 +758,8 @@ CreateBody Conjunction::toGround(ToGroundArg &x, Ground::UStmVec &stms) const {
         return std::move(ret);
     });
 
-    return CreateBody([&completeRef](Ground::ULitVec &lits, bool primary, bool) {
-        if (primary) { lits.emplace_back(gringo_make_unique<Ground::ConjunctionLiteral>(completeRef)); }
+    return CreateBody([&completeRef](Ground::ULitVec &lits, bool primary, bool auxiliary) {
+        if (primary) { lits.emplace_back(gringo_make_unique<Ground::ConjunctionLiteral>(completeRef, auxiliary)); }
     }, std::move(split));
 }
 
@@ -911,6 +917,17 @@ void TupleHeadAggregate::collect(VarTermBoundVec &vars) const {
     }
 }
 
+namespace {
+
+template <class T>
+void zeroLevel(VarTermBoundVec &bound, T const &x) {
+    bound.clear();
+    x->collect(bound, false);
+    for (auto &var : bound) { var.first->level = 0; }
+}
+
+} // namspace
+
 UHeadAggr TupleHeadAggregate::rewriteAggregates(UBodyAggrVec &body) {
     for (auto &x : elems) {
         if (ULit shifted = std::get<1>(x)->shift(false)) {
@@ -919,7 +936,10 @@ UHeadAggr TupleHeadAggregate::rewriteAggregates(UBodyAggrVec &body) {
             std::get<2>(x).emplace_back(std::move(shifted));
         }
     }
+    // NOTE: if it were possible to add further ruleshere,
+    // then also aggregates with more than one element could be supported
     if (elems.size() == 1 && bounds.empty()) {
+        VarTermBoundVec bound;
         auto &elem = elems.front();
         // Note: to handle undefinedness in tuples
         bool weight = fun == AggregateFunction::SUM || fun == AggregateFunction::SUMP;
@@ -927,6 +947,7 @@ UHeadAggr TupleHeadAggregate::rewriteAggregates(UBodyAggrVec &body) {
         Location l = tuple.empty() ? loc() : tuple.front()->loc();
         for (auto &term : tuple) {
             // NOTE: there could be special predicates for is_integer and is_defined
+            zeroLevel(bound, term);
             UTerm first = get_clone(term);
             if (weight) {
                 first = make_locatable<BinOpTerm>(l, BinOp::ADD, std::move(first), make_locatable<ValTerm>(l, Symbol::createNum(0)));
@@ -937,9 +958,11 @@ UHeadAggr TupleHeadAggregate::rewriteAggregates(UBodyAggrVec &body) {
         tuple.clear();
         tuple.emplace_back(make_locatable<ValTerm>(l, Symbol::createNum(0)));
         for (auto &lit : std::get<2>(elem)) {
+            zeroLevel(bound, lit);
             body.emplace_back(gringo_make_unique<SimpleBodyLiteral>(std::move(lit)));
         }
         std::get<2>(elem).clear();
+        zeroLevel(bound, std::get<1>(elem));
     }
     return nullptr;
 }
@@ -1312,7 +1335,16 @@ UHeadAggr Disjunction::rewriteAggregates(UBodyAggrVec &body) {
         }
         if (elem.second.empty() && elem.first.size() == 1) {
             auto &head = elem.first.front();
-            for (auto &lit : head.second) { body.emplace_back(make_locatable<SimpleBodyLiteral>(loc(), std::move(lit))); }
+            VarTermBoundVec vars;
+            head.first->collect(vars, false);
+            for (auto &var : vars) { var.first->level = 0; }
+            vars.clear();
+            for (auto &lit : head.second) {
+                lit->collect(vars, false);
+                for (auto &var : vars) { var.first->level = 0; }
+                vars.clear();
+                body.emplace_back(make_locatable<SimpleBodyLiteral>(loc(), std::move(lit)));
+            }
             head.second.clear();
         }
     }
@@ -1321,16 +1353,29 @@ UHeadAggr Disjunction::rewriteAggregates(UBodyAggrVec &body) {
 
 bool Disjunction::simplify(Projections &project, SimplifyState &state, Logger &log) {
     for (auto &elem : elems) {
-        elem.first.erase(std::remove_if(elem.first.begin(), elem.first.end(), [&](Head &head) {
+        for (auto &head : elem.first) {
+            bool replace = false;
             SimplifyState elemState(state);
-            if (!head.first->simplify(log, project, elemState)) { return true; }
-            for (auto &lit : head.second) {
-                if (!lit->simplify(log, project, elemState)) { return true; }
+            if (head.first->simplify(log, project, elemState)) {
+                for (auto &lit : head.second) {
+                    if (!lit->simplify(log, project, elemState)) {
+                        replace = true;
+                        break;
+                    }
+                }
             }
-            for (auto &dot : elemState.dots) { head.second.emplace_back(RangeLiteral::make(dot)); }
-            for (auto &script : elemState.scripts) { head.second.emplace_back(ScriptLiteral::make(script)); }
-            return false;
-        }), elem.first.end());
+            else { replace = true; }
+            if (replace) {
+                auto loc = head.first->loc();
+                head.first = Gringo::make_locatable<RelationLiteral>(loc, Relation::EQ, make_locatable<ValTerm>(
+                    loc, Symbol::createNum(0)), make_locatable<ValTerm>(loc, Symbol::createNum(0)));
+                head.second.clear();
+            }
+            else {
+                for (auto &dot : elemState.dots) { head.second.emplace_back(RangeLiteral::make(dot)); }
+                for (auto &script : elemState.scripts) { head.second.emplace_back(ScriptLiteral::make(script)); }
+            }
+        }
     }
     elems.erase(std::remove_if(elems.begin(), elems.end(), [&](ElemVec::value_type &elem) -> bool {
         SimplifyState elemState(state);
@@ -1751,8 +1796,8 @@ CreateBody DisjointAggregate::toGround(ToGroundArg &x, Ground::UStmVec &stms) co
             return std::move(ret);
         });
     }
-    return CreateBody([&completeRef, this](Ground::ULitVec &lits, bool primary, bool) {
-        if (primary) { lits.emplace_back(gringo_make_unique<Ground::DisjointLiteral>(completeRef, naf)); }
+    return CreateBody([&completeRef, this](Ground::ULitVec &lits, bool primary, bool auxiliary) {
+        if (primary) { lits.emplace_back(gringo_make_unique<Ground::DisjointLiteral>(completeRef, naf, auxiliary)); }
     }, std::move(split));
 }
 
