@@ -1,20 +1,24 @@
-// {{{ GPL License
+// {{{ MIT License
 
-// This file is part of gringo - a grounder for logic programs.
-// Copyright (C) 2013  Roland Kaminski
+// Copyright 2017 Roland Kaminski
 
-// This program is free software: you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to
+// deal in the Software without restriction, including without limitation the
+// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+// sell copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 
-// You should have received a copy of the GNU General Public License
-// along with this program.  If not, see <http://www.gnu.org/licenses/>.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+// IN THE SOFTWARE.
 
 // }}}
 
@@ -201,7 +205,7 @@ UTheoryTerm TheoryParser::parse(RawTheoryTerm::ElemVec &&elems, Logger &log) {
     for (auto &elem : elems) {
         for (auto &op : elem.first) {
             if (!def_.hasOp(op, unary)) {
-                GRINGO_REPORT(log, clingo_error_runtime)
+                GRINGO_REPORT(log, Warnings::RuntimeError)
                     << loc_ << ": error: missing definition for operator:" << "\n"
                     << "  " << op << "\n";
             }
@@ -235,7 +239,7 @@ RawTheoryTerm &RawTheoryTerm::operator=(RawTheoryTerm &&) = default;
 
 RawTheoryTerm::~RawTheoryTerm() noexcept = default;
 
-void RawTheoryTerm::append(FWStringVec &&ops, UTheoryTerm &&term) {
+void RawTheoryTerm::append(StringVec &&ops, UTheoryTerm &&term) {
     assert(elems_.empty() || !ops.empty());
     elems_.emplace_back(std::move(ops), std::move(term));
 }
@@ -314,7 +318,7 @@ UnaryTheoryTerm *UnaryTheoryTerm::clone() const {
 Potassco::Id_t UnaryTheoryTerm::eval(TheoryData &data, Logger &log) const {
     auto op = data.addTerm(op_.c_str());
     Potassco::Id_t args[] = { arg_->eval(data, log) };
-    return data.addTerm(op, Potassco::toSpan(args, 1));
+    return data.addTermFun(op, Potassco::toSpan(args, 1));
 }
 
 void UnaryTheoryTerm::collect(VarTermBoundVec &vars) {
@@ -360,7 +364,7 @@ BinaryTheoryTerm *BinaryTheoryTerm::clone() const {
 Potassco::Id_t BinaryTheoryTerm::eval(TheoryData &data, Logger &log) const {
     auto op = data.addTerm(op_.c_str());
     Potassco::Id_t args[] = { left_->eval(data, log), right_->eval(data, log) };
-    return data.addTerm(op, Potassco::toSpan(args, 2));
+    return data.addTermFun(op, Potassco::toSpan(args, 2));
 }
 
 void BinaryTheoryTerm::collect(VarTermBoundVec &vars) {
@@ -414,7 +418,7 @@ Potassco::Id_t TupleTheoryTerm::eval(TheoryData &data, Logger &log) const {
     for (auto &arg : args_) {
         args.emplace_back(arg->eval(data, log));
     }
-    return data.addTerm(type_, Potassco::toSpan(args));
+    return data.addTermTup(type_, Potassco::toSpan(args));
 }
 
 void TupleTheoryTerm::collect(VarTermBoundVec &vars) {
@@ -470,7 +474,7 @@ Potassco::Id_t FunctionTheoryTerm::eval(TheoryData &data, Logger &log) const {
     for (auto &arg : args_) {
         args.emplace_back(arg->eval(data, log));
     }
-    return data.addTerm(name, Potassco::toSpan(args));
+    return data.addTermFun(name, Potassco::toSpan(args));
 }
 
 void FunctionTheoryTerm::collect(VarTermBoundVec &vars) {
@@ -565,11 +569,11 @@ Potassco::Id_t TheoryData::addTerm(char const *name) {
     return addTerm_(name);
 }
 
-Potassco::Id_t TheoryData::addTerm(Potassco::Id_t funcSym, Potassco::IdSpan const& terms) {
+Potassco::Id_t TheoryData::addTermFun(Potassco::Id_t funcSym, Potassco::IdSpan const& terms) {
     return addTerm_(funcSym, terms);
 }
 
-Potassco::Id_t TheoryData::addTerm(Potassco::Tuple_t type, Potassco::IdSpan const& terms) {
+Potassco::Id_t TheoryData::addTermTup(Potassco::Tuple_t type, Potassco::IdSpan const& terms) {
     return addTerm_(type, terms);
 }
 
@@ -581,7 +585,7 @@ Potassco::Id_t TheoryData::addTerm(Symbol value) {
                 auto f = addTerm("-");
                 auto ret = addTerm(-num);
                 Potassco::Id_t args[] = { ret };
-                return addTerm(f, Potassco::toSpan(args, 1));
+                return addTermFun(f, Potassco::toSpan(args, 1));
             }
             else {
                 return addTerm(num);
@@ -606,16 +610,16 @@ Potassco::Id_t TheoryData::addTerm(Symbol value) {
                 args.emplace_back(addTerm(arg));
             }
             if (value.name().empty()) {
-                return addTerm(Potassco::Tuple_t::Paren, Potassco::toSpan(args));
+                return addTermTup(Potassco::Tuple_t::Paren, Potassco::toSpan(args));
             }
             else {
                 Potassco::Id_t name = addTerm(value.name().c_str());
                 auto ret = args.empty()
                     ? addTerm(value.name().c_str())
-                    : addTerm(name, Potassco::toSpan(args));
+                    : addTermFun(name, Potassco::toSpan(args));
                 if (value.sign()) {
                     Potassco::Id_t f = addTerm("-");
-                    ret = addTerm(f, Potassco::toSpan(&ret, 1));
+                    ret = addTermFun(f, Potassco::toSpan(&ret, 1));
                 }
                 return ret;
             }
@@ -638,7 +642,7 @@ Potassco::Id_t TheoryData::addElem(Potassco::IdSpan const &tuple, LitVec &&cond)
     }, [&](Potassco::Id_t const &a, Potassco::Id_t const &b) {
         assert(a < size);
         return b == size ? elementEqual(data_.getElement(a), conditions_[a], tuple, cond) : a == b;
-    }, conditions_.size());
+    }, numeric_cast<unsigned>(conditions_.size()));
     if (ret.second) {
         data_.addElement(size, tuple, cond.empty() ? 0 : Potassco::TheoryData::COND_DEFERRED);
         conditions_.emplace_back(std::move(cond));
@@ -739,11 +743,12 @@ void TheoryData::setCondition(Potassco::Id_t elementId, Potassco::Id_t newCond)
     data_.setCondition(elementId, newCond);
 }
 
-void TheoryData::reset() {
+void TheoryData::reset(bool resetData) {
     TIdSet().swap(terms_);
     TIdSet().swap(elems_);
     AtomSet().swap(atoms_);
     ConditionVec().swap(conditions_);
+    if (resetData) { data_.reset(); }
 }
 
 bool TheoryData::hasConditions() const {
