@@ -1,20 +1,24 @@
-// {{{ GPL License
+// {{{ MIT License
 
-// This file is part of gringo - a grounder for logic programs.
-// Copyright (C) 2013  Roland Kaminski
+// Copyright 2017 Roland Kaminski
 
-// This program is free software: you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to
+// deal in the Software without restriction, including without limitation the
+// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+// sell copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 
-// You should have received a copy of the GNU General Public License
-// along with this program.  If not, see <http://www.gnu.org/licenses/>.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+// IN THE SOFTWARE.
 
 // }}}
 
@@ -22,7 +26,6 @@
 #include "gringo/input/program.hh"
 #include "gringo/ground/program.hh"
 #include "gringo/output/output.hh"
-#include "gringo/scripts.hh"
 
 #include "tests/tests.hh"
 
@@ -45,14 +48,15 @@ std::string ground(std::string const &str, std::initializer_list<std::string> fi
     Input::Program prg;
     Defines defs;
     Gringo::Test::TestGringoModule module;
-    Scripts scripts(module);
-    Input::NongroundProgramBuilder pb{ scripts, prg, out, defs };
-    Input::NonGroundParser ngp{ pb };
+    Gringo::Test::TestContext context;
+    Input::NongroundProgramBuilder pb{ context, prg, out, defs };
+    bool incmode;
+    Input::NonGroundParser ngp{ pb, incmode };
     ngp.pushStream("-", gringo_make_unique<std::stringstream>(str), module);
     ngp.parse(module);
     prg.rewrite(defs, module);
     Program gPrg(prg.toGround(out.data, module));
-    gPrg.ground(scripts, out, module);
+    gPrg.ground(context, out, module);
 
     std::string line;
     std::vector<std::string> res;
@@ -1108,6 +1112,14 @@ TEST_CASE("ground-instantiation", "[ground]") {
                 "-q(X):-not not -q(X), p(X).\n"
                 " q(X):-not not  q(X), p(X).\n"));
     }
+    SECTION("two aggregates") {
+        REQUIRE(
+            "p:-#count{0,a:a}=0.\n"
+            "p:-#count{0,a:a}=1,1<=#count{0,b:b}.\n"
+            "{a;b}.\n" == ground(
+                "{a;b}.\n"
+                "p :- X = { a }, X { b }.\n"));
+    }
 
     SECTION("tuple") {
         REQUIRE("p(((),())).\n" == ground("p(((),())).\n"));
