@@ -1,20 +1,24 @@
-// {{{ GPL License
+// {{{ MIT License
 
-// This file is part of gringo - a grounder for logic programs.
-// Copyright (C) 2013  Roland Kaminski
+// Copyright 2017 Roland Kaminski
 
-// This program is free software: you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to
+// deal in the Software without restriction, including without limitation the
+// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+// sell copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 
-// You should have received a copy of the GNU General Public License
-// along with this program.  If not, see <http://www.gnu.org/licenses/>.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+// IN THE SOFTWARE.
 
 // }}}
 
@@ -26,6 +30,9 @@
 #include <climits>
 #include <sstream>
 #include <functional>
+#ifdef _MSC_VER
+#pragma warning (disable : 4503) // decorated name length exceeded
+#endif
 
 namespace Gringo { namespace Test {
 
@@ -171,7 +178,7 @@ TEST_CASE("term", "[base]") {
         REQUIRE(*gterm(lin("X",2,3)) != *gterm(lin("X",1,3)));
         REQUIRE(*gterm(lin("X",2,3)) != *gterm(lin("X",2,2)));
         REQUIRE(*gterm(var("X")) == *gterm(var("X")));
-        REQUIRE(*gterm(fun("f", var("X"), var("X"))) == *gterm(fun("f", var("X"), var("X", 1)))); // Note: all levels=0
+        REQUIRE(*gterm(fun("f", var("X"), var("X"))) != *gterm(fun("f", var("X"), var("X", 1)))); // Note: all levels=0
         REQUIRE(*gterm(var("X")) == *gterm(var("Y"))); // Note: intended
         REQUIRE(*gterm(fun("f", val(NUM(1)), val(NUM(2)))) == *gterm(fun("f", val(NUM(1)), val(NUM(2)))));
         REQUIRE(*gterm(fun("f", val(NUM(1)), val(NUM(2)))) != *gterm(fun("g", val(NUM(1)), val(NUM(2)))));
@@ -252,13 +259,20 @@ TEST_CASE("term", "[base]") {
 
     SECTION("undefined") {
         bool undefined = false;
+
         REQUIRE(NUM(0) == binop(BinOp::POW, val(ID("a")), val(NUM(1)))->eval(undefined, log));
         REQUIRE(undefined);
         REQUIRE("dummy:1:1: info: operation undefined:\n  (a**1)\n" == log.messages().back());
+
         undefined = false;
         REQUIRE(NUM(0) == unop(UnOp::NOT, val(ID("a")))->eval(undefined, log));
         REQUIRE(undefined);
         REQUIRE("dummy:1:1: info: operation undefined:\n  (~a)\n" == log.messages().back());
+
+        undefined = false;
+        REQUIRE(NUM(0) == binop(BinOp::POW, val(NUM(0)), val(NUM(-2)))->eval(undefined, log));
+        REQUIRE(undefined);
+        REQUIRE("dummy:1:1: info: operation undefined:\n  (0**-2)\n" == log.messages().back());
     }
 
     SECTION("project") {
